// ball.cc - functions required by enemy balls
//
// Copyright (C) 2000 Trevor Spiteri
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "ball.hh"
#include "misc.hh"
#include "record.hh"

void balls::enemy::move(int t)
{
    // Modify vel because of collisions.
    v += dv;
    dv.set();
      
    // Add velocity * time to distance
    l += v * t;
      
    if (hit_wall(this)) {
        bounce_wall(this);
        get_inside_wall(this);
    }
      
    if (strength <= 0)
        rem = true;
}

void balls::enemy::do_collide(const ball* other)
{
    // enemy balls are not damaged when bouncing against each other
    if (!is_enemy(other)) {
        int d = other -> damage();
        if (hardness < 1000)
            strength -= d * (1000 - hardness) / 1000;
        hardness -= d;
        if (hardness < 0)
            hardness = 0;
        if (strength <= 0) {
            const ship_bullet* sb = dynamic_cast<const ship_bullet*>(other);
            if (sb)
                sb->add_score(score);
        }
    }
    dv += dvel_collision(this, other);
}

int balls::enemy::damage() const
{
    return mas;
}

balls::normal_ball::normal_ball(vect3d<> ll, vect3d<> vv, draw_ball_info* inf)
    : enemy((*inf->bd)["normal_ball"]["radius"],
            (*inf->bd)["normal_ball"]["mass"],
            ll, vv, inf,
            (*inf->bd)["normal_ball"]["strength"],
            (*inf->bd)["normal_ball"]["hardness"],
            (*inf->bd)["normal_ball"]["score"]),
      graphics::deep_drawable((*inf->gallery)["normal_ball"],
                              inf->mapp)
{
}

void balls::normal_ball::move(int t)
{
    enemy::move(t);
    if (strength <= 0) {
        rec::record& r = (*info->bd)["normal_ball"];
        int frags = r["frags"];
        for (int i = 0; i < frags; ++i) {
            vect3d<> floc(l), fvel(v);
            floc += rnd3d(*info->rand, double(r["radius"]) * 2);
            fvel += rnd3d(*info->rand, double(r["velxy"]));
            ball* frag = new normal_frag(floc, fvel,
                                         static_cast<draw_ball_info*>(info));
            info->swal -> swallow(frag);
        }
    }
}

balls::normal_frag::normal_frag(vect3d<> ll, vect3d<> vv, draw_ball_info* inf)
    : bullet((*inf->bd)["normal_frag"]["radius"],
             (*inf->bd)["normal_frag"]["mass"],
             ll, vv, inf,
             (*inf->bd)["normal_frag"]["damage"]),
      graphics::deep_drawable((*inf->gallery)["normal_frag"],
                              inf->mapp)
{
}

balls::hard_ball::hard_ball(vect3d<> ll, vect3d<> vv, draw_ball_info* inf)
    : enemy((*inf->bd)["hard_ball"]["radius"],
            (*inf->bd)["hard_ball"]["mass"],
            ll, vv, inf,
            (*inf->bd)["hard_ball"]["strength"],
            (*inf->bd)["hard_ball"]["hardness"],
            (*inf->bd)["hard_ball"]["score"]),
      graphics::deep_drawable((*inf->gallery)["hard_ball"],
                              inf->mapp)
{
}

void balls::hard_ball::move(int t)
{
    enemy::move(t);
    if (strength <= 0) {
        rec::record& r = (*info->bd)["hard_ball"];
        int frags = r["frags"];
        for (int i = 0; i < frags; ++i) {
            vect3d<> floc(l), fvel(v);
            floc += rnd3d(*info->rand, double(r["radius"]) * 2);
            fvel += rnd3d(*info->rand, double(r["velxy"]));
            ball* frag = new hard_frag(floc, fvel,
                                       static_cast<draw_ball_info*>(info));
            info->swal -> swallow(frag);
        }
    }
}

balls::hard_frag::hard_frag(vect3d<> ll, vect3d<> vv, draw_ball_info* inf)
    : bullet((*inf->bd)["hard_frag"]["radius"],
             (*inf->bd)["hard_frag"]["mass"],
             ll, vv, inf,
             (*inf->bd)["hard_frag"]["damage"]),
      graphics::deep_drawable((*inf->gallery)["hard_frag"],
                              inf->mapp)
{
}

balls::fast_ball::fast_ball(vect3d<> ll, vect3d<> vv, draw_ball_info* inf)
    : enemy((*inf->bd)["fast_ball"]["radius"],
            (*inf->bd)["fast_ball"]["mass"],
            ll, vv, inf,
            (*inf->bd)["fast_ball"]["strength"],
            (*inf->bd)["fast_ball"]["hardness"],
            (*inf->bd)["fast_ball"]["score"]),
      graphics::deep_drawable((*inf->gallery)["fast_ball"],
                              inf->mapp)
{
}

void balls::fast_ball::move(int t)
{
    enemy::move(t);
    if (strength <= 0) {
        rec::record& r = (*info->bd)["fast_ball"];
        int frags = r["frags"];
        for (int i = 0; i < frags; ++i) {
            vect3d<> floc(l), fvel(v);
            floc += rnd3d(*info->rand, double(r["radius"]) * 2);
            fvel += rnd3d(*info->rand, double(r["velxy"]));
            ball* frag = new fast_frag(floc, fvel,
                                       static_cast<draw_ball_info*>(info));
            info->swal -> swallow(frag);
        }
    }
}

balls::fast_frag::fast_frag(vect3d<> ll, vect3d<> vv, draw_ball_info* inf)
    : bullet((*inf->bd)["fast_frag"]["radius"],
             (*inf->bd)["fast_frag"]["mass"],
             ll, vv, inf,
             (*inf->bd)["fast_frag"]["damage"]),
      graphics::deep_drawable((*inf->gallery)["fast_frag"],
                              inf->mapp)
{
}

balls::big_ball::big_ball(vect3d<> ll, vect3d<> vv, draw_ball_info* inf)
    : enemy((*inf->bd)["big_ball"]["radius"],
            (*inf->bd)["big_ball"]["mass"],
            ll, vv, inf,
            (*inf->bd)["big_ball"]["strength"],
            (*inf->bd)["big_ball"]["hardness"],
            (*inf->bd)["big_ball"]["score"]),
      graphics::deep_drawable((*inf->gallery)["big_ball"],
                              inf->mapp)
{
}

void balls::big_ball::move(int t)
{
    enemy::move(t);
    if (strength <= 0) {
        rec::record& r = (*info->bd)["big_ball"];
        int frags = r["frags"];
        for (int i = 0; i < frags; ++i) {
            vect3d<> floc(l), fvel(v);
            floc += rnd3d(*info->rand, double(r["radius"]) * 2);
            fvel += rnd3d(*info->rand, double(r["velxy"]));
            ball* frag = new big_frag(floc, fvel,
                                      static_cast<draw_ball_info*>(info));
            info->swal -> swallow(frag);
        }
    }
}

balls::big_frag::big_frag(vect3d<> ll, vect3d<> vv, draw_ball_info* inf)
    : bullet((*inf->bd)["big_frag"]["radius"],
             (*inf->bd)["big_frag"]["mass"],
             ll, vv, inf,
             (*inf->bd)["big_frag"]["damage"]),
      graphics::deep_drawable((*inf->gallery)["big_frag"],
                              inf->mapp)
{
}

balls::whiz_ball::whiz_ball(vect3d<> ll, vect3d<> vv, draw_ball_info* inf)
    : enemy((*inf->bd)["whiz_ball"]["radius"],
            (*inf->bd)["whiz_ball"]["mass"],
            ll, vv, inf,
            (*inf->bd)["whiz_ball"]["strength"],
            (*inf->bd)["whiz_ball"]["hardness"],
            (*inf->bd)["whiz_ball"]["score"]),
      graphics::deep_drawable((*inf->gallery)["whiz_ball"],
                              inf->mapp),
      time_change((*inf->bd)["whiz_ball"]["timechange"]),
      time_left(time_change)
{
}

void balls::whiz_ball::move(int t)
{
    enemy::move(t);
    time_left -= t;
    if (time_left <= 0) {
        do {
            time_left += time_change;
        } while (time_left <= 0);
        double r = misc::mag(v);
        double a = (*info->rand)(M_PI);
        v = misc::mkpolar(r, a);
    }
    if (strength <= 0) {
        rec::record& r = (*info->bd)["whiz_ball"];
        int frags = r["frags"];
        for (int i = 0; i < frags; ++i) {
            vect3d<> floc(l), fvel(v);
            floc += rnd3d(*info->rand, double(r["radius"]) * 2);
            fvel += rnd3d(*info->rand, double(r["velxy"]));
            ball* frag = new whiz_frag(floc, fvel,
                                       static_cast<draw_ball_info*>(info));
            info->swal -> swallow(frag);
        }
    }
}

balls::whiz_frag::whiz_frag(vect3d<> ll, vect3d<> vv, draw_ball_info* inf)
    : bullet((*inf->bd)["whiz_frag"]["radius"],
             (*inf->bd)["whiz_frag"]["mass"],
             ll, vv, inf,
             (*inf->bd)["whiz_frag"]["damage"]),
      graphics::deep_drawable((*inf->gallery)["whiz_frag"],
                              inf->mapp)
{
}

balls::normal_bullet::normal_bullet(vect3d<> ll, vect3d<> vv,
                                    draw_ball_info* inf, ship* s)
    : ship_bullet((*inf->bd)["normal_bullet"]["radius"],
                  (*inf->bd)["normal_bullet"]["mass"],
                  ll, vv, inf,
                  (*inf->bd)["normal_bullet"]["damage"], s),
      graphics::deep_drawable((*inf->gallery)["normal_bullet"],
                              inf->mapp)
{
}
