// game.cc - functions having to do with the game
//
// Copyright (C) 2000 Trevor Spiteri
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <algorithm>
#include <fstream>
#include <strstream>
#include <functional>
#include <iterator>
#include <list>

#include "ball.hh"
#include "game.hh"
#include "new_ball.hh"

// Coversion factors of velocity and depth from double to int
const double Game::game::velconv = 200;
const double Game::game::depthconv = 4 / tunnel_d;

Game::cpanel::cpanel(graphics::tsurface* t, int x, int y, int w, int h,
                     int vmin, int vmax, int smin, int smax,
                     int hmin, int hmax, int amin, int amax,
                     const rec::record& r)
    : score(0), depth(0), time(0),
      aminv(vmin), amaxv(vmax), amins(smin), amaxs(smax),
      aminh(hmin), amaxh(hmax), amina(amin), amaxa(amax),
      ts(t), X(x), Y(y)
{
    graphics::raw_image* rimages[itotal];
    rimages[0] = ts->make_raw_image(r["0"]);
    rimages[1] = ts->make_raw_image(r["1"]);
    rimages[2] = ts->make_raw_image(r["2"]);
    rimages[3] = ts->make_raw_image(r["3"]);
    rimages[4] = ts->make_raw_image(r["4"]);
    rimages[5] = ts->make_raw_image(r["5"]);
    rimages[6] = ts->make_raw_image(r["6"]);
    rimages[7] = ts->make_raw_image(r["7"]);
    rimages[8] = ts->make_raw_image(r["8"]);
    rimages[9] = ts->make_raw_image(r["9"]);
    rimages[iscore] = ts->make_raw_image(r["score"]);
    rimages[idepth] = ts->make_raw_image(r["depth"]);
    rimages[itime] = ts->make_raw_image(r["time"]);
    rimages[ivel] = ts->make_raw_image(r["velocity"]);
    rimages[istr] = ts->make_raw_image(r["strength"]);
    rimages[ihard] = ts->make_raw_image(r["hardness"]);
    rimages[iammo] = ts->make_raw_image(r["ammo"]);

    unitx = rimages[0] -> w();
    unity = rimages[1] -> h();
    int sep = unity / 2;
    int isep = unity;
    int ssep = sep + isep;

    ys[0] = 0; // SCORE
    ys[1] = ys[0] + rimages[iscore]->h() + sep; // number
    ys[2] = ys[1] + unity + isep; // DEPTH
    ys[3] = ys[2] + rimages[idepth]->h() + sep; // number
    ys[4] = ys[3] + unity + isep; // TIME
    ys[5] = ys[4] + rimages[itime]->h() + sep; // number
    ys[6] = ys[5] + unity + ssep; // VELOCITY
    ys[7] = ys[6] + rimages[ivel]->h() + sep; // dial
    ys[8] = ys[7] + unity + isep; // STRENGTH
    ys[9] = ys[8] + rimages[istr]->h() + sep; // dial
    ys[10] = ys[9] + unity + isep; // HARDNESS
    ys[11] = ys[10] + rimages[ihard]->h() + sep; // dial
    ys[12] = ys[11] + unity + isep; // AMMO
    ys[13] = ys[12] + rimages[iammo]->h() + sep; // dial
    ys[14] = ys[13] + unity; // bottom

    int rw = 9 * unitx;
    for (int i = 10; i < itotal; ++i)
        if (rimages[i]->w() > rw)
            rw = rimages[i]->w();
    int rh = ys[draw_items];

    // if rw/w > rh/h
    if (rw * h > rh * w) {
        int newh = rh * w / rw;
        Y += (h - newh) / 2;
        h = newh;
    }
    else {
        int neww = rw * h / rh;
        X += (w - neww) / 2;
        w = neww;
    }

    misc::scaler sc(rw, rh, w, h);
    for (int i = 0; i <= draw_items; ++i)
        ys[i] = Y + sc.geth(ys[i]);
    unitx = sc.getw(unitx);
    unity = sc.geth(unity);

    for (int i = 0; i < itotal; ++i) {
        images[i] = ts->make_image(rimages[i], sc);
        delete rimages[i];
    }

    graphics::raw_image* rdial = ts->make_raw_image(r["dial"]);
    dials[dvel] = ts->make_dial(rdial, vmin, vmax, 9*unitx, unity);
    dials[dstr] = ts->make_dial(rdial, smin, smax, 9*unitx, unity);
    dials[dhard] = ts->make_dial(rdial, hmin, hmax, 9*unitx, unity);
    dials[dammo] = ts->make_dial(rdial, amin, amax, 9*unitx, unity);
    delete rdial;
}

Game::cpanel::~cpanel()
{
    for (int i = 0; i < itotal; ++i)
        delete images[i];
}

void Game::cpanel::draw() const
{
    images[iscore]->draw(X, ys[0]);
    draw_num(score, X, ys[1]);
    images[idepth]->draw(X, ys[2]);
    draw_num(depth, X, ys[3]);
    images[itime]->draw(X, ys[4]);
    draw_num(time, X, ys[5]);
    images[ivel]->draw(X, ys[6]);
    dials[dvel]->draw(X, ys[7], velocity, cminv, cmaxv);
    images[istr]->draw(X, ys[8]);
    dials[dvel]->draw(X, ys[9], strength, cmins, cmaxs);
    images[ihard]->draw(X, ys[10]);
    dials[dhard]->draw(X, ys[11], hardness, cminh, cmaxh);
    images[iammo]->draw(X, ys[12]);
    dials[dammo]->draw(X, ys[13], ammo, cmina, cmaxa);
}

void Game::cpanel::draw_num(int n, int x, int y) const
{
    int digit[20];
    int len = 0;
    if (n < 0)
        n = 0;
    do {
        digit[len] = n % 10;
        n /= 10;
        ++len;
    } while(n);

    while (--len >= 0) {
        images[digit[len]]->draw(x, y);
        x += unitx;
    }
}

//  void Game::pauser::handle(int)
//  {
//      ga->pause_game();
//  }

Game::game::game(int xr, int yr, const rec::record& options,
                 graphics::tsurface* ts, rec::recrecord* b_d,
                 control::controller* c, Timer::timer* t, misc::random& r)
    : xres(xr), yres(yr),
      Time_passed(0), Depth_moved(0),
      cp(ts, getx(options["cpanel_x"]), gety(options["cpanel_y"]),
         getx(options["cpanel_w"]), gety(options["cpanel_h"]),
         intvel((*b_d)["the_ship"]["vmin"]),
         intvel((*b_d)["the_ship"]["vmax"]),
         (*b_d)["the_ship"]["smin"], (*b_d)["the_ship"]["smax"],
         (*b_d)["the_ship"]["hmin"], (*b_d)["the_ship"]["hmax"],
         (*b_d)["the_ship"]["amin"], (*b_d)["the_ship"]["amax"],
         options["cpanel_images"].ref(rec::tag<rec::record>())),
      wicon_x(getx(options["weapon_x"])),
      wicon_y(gety(options["weapon_y"])),
      Stat(start),                         // indicates game is just starting
      Cycle_dur(options["cycle"]),
      Timer_cycle_dur(options["timer_cycle"]),
      max_time_wo_draw(options["maxtime_nodraw"]),
      time_wo_ships(options["time_wo_ships"]),
      time_left(time_wo_ships),
      rnd(r),
      bd(b_d),
      where_to_draw(ts),
      proj(tunnel_d, tunnel_d, tunnel_r,
           getx(options["width"]), gety(options["height"])),
      mapp(proj, getx(options["center_x"]), gety(options["center_y"])),
      cont(c),
      tmr(t),
      tun(ts->make_tunnel(&mapp))
{
    // fill up gallery
    int wiw = getx(options["weapon_w"]);
    int wih = gety(options["weapon_h"]);
    for (rec::recrecord::iterator i = bd->begin(); i != bd->end(); ++i) {
        rec::record& r = i->second;
        if (r.find("filename") != r.end()) {
            double dia = 2 * double(r["radius"]);
            graphics::deep_image* d =
                ts->make_deep_image(ts->make_mask_raw_image(r["filename"]),
                                    32, proj, dia, dia);
            gallery[i->first] = d;
        }
        if (r.find("icon") != r.end()) {
            graphics::raw_image* ri = ts->make_raw_image(r["icon"]);
            misc::scaler sc(ri->w(), ri->h(), wiw, wih);
            graphics::image* im = ts->make_image(ri, sc);
            wicons[i->first] = im;
        }
    }
    
    swallower = balls::make_swallower(front_inserter(flyers));
    
    db_info.bd = bd;
    db_info.swal = swallower;
    db_info.rand = &rnd;
    db_info.gallery = &gallery;
    db_info.wicons = &wicons;
    db_info.mapp = &mapp;

    // First of all, we create the ship and the master_ball.
    balls::the_ship* thes = new balls::the_ship(misc::vect3d<>(0, 0, 0),
        misc::vect3d<>(0, 0, (*db_info.bd)["the_ship"]["minvelz"]),
        &db_info, cont);
    zero_depth_ball = thes;
    crea = new balls::def_creator(&db_info, thes);
    flyers.push_back(thes);
    ships.push_back(thes);

//      std::ifstream is("data/new_ball.dat");
//      flyers.push_back(new balls::master_ball(is, &db_info,
//                           static_cast<balls::ship*>(zero_depth_ball)));
}

namespace {
    struct del_ball {
        void operator()(balls::ball* b) { delete b; }
    };

    struct del_di {
        void operator()(std::pair<const std::string, graphics::deep_image*> p)
        { delete p.second; }
    };
    
    struct del_i {
        void operator()(std::pair<const std::string, graphics::image*> p)
        { delete p.second; }
    };
    
    int frames = 0;
}

Game::game::~game()
{
    std::for_each(flyers.begin(), flyers.end(), del_ball());
    delete swallower;
    std::for_each(wicons.begin(), wicons.end(), del_i());
    std::for_each(gallery.begin(), gallery.end(), del_di());
    delete cont;
}

// start the game until it is paused
void Game::game::play()
{
    // make sure the game is not over yet
    if (Stat == done)
        return;
   
    // set Stat to run in case the game was paused or just starting
    Stat = run;
   
    bool draw_now = true; // false when no changes have occured
   
   // start the timer and set the count to 0 before starting
    tmr -> set_delay(Timer_cycle_dur * unit_msec);
    tmr -> start();
   
   // main game loop
    while (Stat == run) {
        // draw the game section if something has changed
        if (draw_now) {
            draw_now = false;
            draw();
            ++frames;
            time_since_last_draw = 0;
        }

        tmr -> wait();

        // check if we need to pause
        cont->poll();
        if (cont->cont1()) {
            Stat = pause;
            break;
        }
      
        Time_passed += Cycle_dur;
        time_since_last_draw += Cycle_dur;
        draw_now = true;  // we are going to change everything
      
        // Every item must make a move. In the move, if a ball
        // is broken up, its remove flag is set.
        move_flyers();
      
        // Then we sort them according to z coordinate. Note that before
        // sorting, broken up balls are removed.
        sort_flyers();
      
        // Make sure that the zero_depth_ship has depth 0, and move the
        // other balls accordingly so that there is no relative movement.
        // This should also update depth.
        // Also check if there are any collisions between balls and
        // inform concerned balls.
        check_flyers_depth_and_collisions();
      
        // See if we are to go on. If there aren't any ships, start
        // decrementing time_left until we can go out.
        if (ships.empty() && (time_left -= Cycle_dur) <= 0)
            Stat = done;

        tmr -> done();

        // If it has NOT been a long time since we have last drawn,
        // and we are lagging in time, skip draw.
        if (time_since_last_draw < max_time_wo_draw
            && tmr -> ticks() > 0) {
            draw_now = false;
        }
    }
   
    // Before exiting, stop timer.
    tmr -> stop();
}

void Game::game::draw()
{
    where_to_draw->clear(0, 0, 0);
    tun->draw(misc::vect3d<>(0, 0, 0));
    balls::the_ship* s
        = static_cast<balls::the_ship*>(zero_depth_ball);
    cp.score = s->score;
    cp.depth = intdepth(Depth_moved);
    cp.time = Time_passed / 1000;
    cp.velocity = intvel(s->vel().z());
    cp.cminv = intvel(s->min_zvel);
    cp.cmaxv = intvel(s->max_zvel);
    cp.strength = s->strength;
    cp.cmins = 0;
    cp.cmaxs = cp.amaxs;
    cp.hardness = s->hardness;
    cp.cminh = 0;
    cp.cmaxh = cp.amaxh;
    cp.ammo = s->ammo;
    cp.cmina = 0;
    cp.cmaxa = cp.amaxa;
    cp.draw();
    (*s->cur_weap)->get_image()->draw(wicon_x, wicon_y);

    // first draw tunnel
      
    // then draw objects
    // We start from the end as the deepest balls are at the end of
    // the list.
    std::for_each(flyers.rbegin(), flyers.rend(), balls::do_ball_draw());

    where_to_draw->flip();
}

void Game::game::move_flyers()
{
    cont->poll();
    for (std::list<ball*>::iterator i = flyers.begin();
         i != flyers.end(); ++i) {
        (*i)->move(Cycle_dur);
    }
    crea->move(Cycle_dur);
}

void Game::game::sort_flyers()
{
    // first remove removables in, then sort, both ships and flyers

    // when we remove from ships, there is still a pointer to the ball
    //    in flyers, so we do not delete it from memory
    ships.remove_if(std::mem_fun(&(ball::remove)));
    ships.sort(balls::above);
      
    for (std::list<ball*>::iterator i = flyers.begin();
         i != flyers.end(); ) {
        double d = (*i)->loc().z();
        if ((*i) -> remove()
            || d < mapp.get_depth() - tunnel_r
            || d > mapp.get_depth() + tunnel_dpth + tunnel_r) {
            if (*i != zero_depth_ball)
                delete *i;
            i = flyers.erase(i);
        }
        else
            ++i;
    }
    flyers.sort(balls::above);
}

void Game::game::check_flyers_depth_and_collisions()
{
    if (flyers.empty())
        return;
      
    // All balls, (including zero_depth_ball), will be moved up by 
    // zero_depth_ball -> depth().
    // Then, check which balls below them they collide with

    double depth = zero_depth_ball -> loc().z();
    Depth_moved = depth;
      
    tun->move(depth);
    mapp.set_depth(depth);
      
    // We have to check for collisions with the ball at first.
    // If there are no more balls (1st condition of nested for),
    // or the remaining balls are all well below the ball at first
    // (2nd condition), stop.
    for (std::list<ball*>::iterator i = flyers.begin();
         i != flyers.end(); ++i) {
        std::list<ball*>::iterator other = i;
        for (++other;
             other != flyers.end() && !balls::well_below(*other, *i);
             ++other) {
            if (balls::collide(*other, *i)) {
                (*other) -> do_collide(*i);
                (*i) -> do_collide(*other);
            }
        }
         
    }
}
