// graphics.hh - interface for graphics and other miscellany
//
// Copyright (C) 2000 Trevor Spiteri
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#ifndef GRAPHICS_HH
#define GRAPHICS_HH

#include <cstddef>
#include <map>
#include <string>
#include <vector>

#include "misc.hh"

namespace graphics {

    // abstract base class image
    // 
    // Stores an image
    class raw_image {
    public:
        virtual ~raw_image() { }

        virtual int w() const = 0;
        virtual int h() const = 0;
    };

    // abstract base class image
    // 
    // Stores an image
    class image {
    public:
        virtual ~image() { }

        virtual void draw(int x, int y) const = 0;
    };

    // abstract base class deep_image
    // 
    // Stores a number of images which can be drawn given a location
    // and a mapper which gives the x,y coords of the surface on which
    // to draw from the location.
    // The implementations should keep track of the target on which they
    // should be drawn.
    class deep_image {
    public:
        virtual ~deep_image() { }

        virtual void draw(const misc::vect3d<>& loc,
                          const misc::mapper<>& m) const = 0;
    };

    // abstract base class dial
    class dial {
    public:
        virtual ~dial() { }

        virtual void draw(int x, int y, int actual,
                          int min, int max) const = 0;
    };

    // a (deep_)image_map is a map containing (deep_)images with a name key
    typedef std::map<const std::string, image*>
        image_map;
    typedef std::map<const std::string, deep_image*>
        deep_image_map;
    
    // abstract base class tsurface
    //    Target surface on which to draw balls, etc..
    // virtual functions:
    //    flip():               update the display
    //    w(), h():             get the width/height
    //    clear(...):           clear the surface with the given colour
    //    make_deep_image(...): make a deep image from a filename. The
    //                          deep_image draws itself onto this tsurface.
    class tunnel_drawable;
    class tsurface {
    public:
        virtual ~tsurface() { }

        virtual void flip() = 0;
        virtual int w() const = 0;
        virtual int h() const = 0;
        virtual void clear(int r, int g, int b) = 0;
        virtual raw_image* make_raw_image(const std::string& filename) = 0;
        virtual raw_image* make_mask_raw_image(const std::string& filename)
            = 0;
        virtual image* make_image(const raw_image* src) = 0;
        virtual image* make_image(const raw_image* src,
                                  const misc::scaler& sc) = 0;
        virtual deep_image* make_deep_image(const raw_image* src, int n,
                                            const misc::projector<>& p,
                                            double w, double h) = 0;
        virtual tunnel_drawable* make_tunnel(const misc::mapper<>* m) = 0;
        virtual dial* make_dial(const raw_image* src, int amin, int amax,
                                int w, int h) = 0;
    };

    // a tsurface* and a mapper*
    class tmsurface {
    public:
        tmsurface(tsurface* t, const misc::mapper<>* m)
            : ts(t), mp(m) { }

        tsurface* get_ts() const { return ts; }
        const misc::mapper<>* get_mp() const { return mp; }
    private:
        tsurface* ts;
        const misc::mapper<>* mp;
    };

    // abstract base class drawable
    //    Something which can be drawn given its location
    class drawable {
    public:
        virtual ~drawable() { }
        virtual void draw(const misc::vect3d<>& loc) const = 0;
    };

    // A drawable made from a deep_image. This remembers the mapper
    // and draws the deep_image when required.
    class deep_drawable : public drawable {
    public:
        deep_drawable(const deep_image* i, const misc::mapper<>* m)
            : di(i), mp(m) { }

        void draw(const misc::vect3d<>& loc) const
        {
            di->draw(loc, *mp);
        }
    protected:
        const deep_image* di;
        const misc::mapper<>* mp;
    };

    class tunnel_drawable : public drawable {
    public:
        virtual void move(double d) = 0;
    };

    // class ellipse
    //    gives points which are inside or exactly on the quadrant of the
    //    ellipse of radii Rx, Ry. Starts with point (0, Ry) and ends with
    //    point (Rx, 0).
    // member functions
    //    next(): goes to next point, returns false if there isn't
    //    x():    gives current x
    //    y():    gives current y
    // NOTE
    //    Implementation is correct for radii up to about 1000 (a little more
    //    in fact).
    class ellipse {
    public:
        ellipse()               { reset(0, 0); }
        ellipse(int Rx, int Ry) { reset(Rx, Ry); }

        void reset(int Rx, int Ry);
        void reset()            { reset(rx, ry); }
        
        bool next();

        int x() const { return cx; }
        int y() const { return cy; }
    private:
        int rx, ry, cx, cy;
        int f, dfx, dfy, ddfx, ddfy;
        bool second_stage;
    };

    // class line_ellipse
    //    gives the amount of pixels needed to draw each horizontal line
    //    of an ellipse of radii rx, ry. x() gives the number of pixels
    //    and y() gives which line. Starts with (0, ry) indicating that
    //    a line ry up should not be drawn and ends with (rx, 0) so that
    //    in the bottom line, rx pixels are drawn.
    // member functions
    //    next(): goes to next line, returns false if there isn't
    //    x():    gives current x
    //    y():    gives current y
    class line_ellipse {
    public:
        line_ellipse()               { reset (0, 0); }
        line_ellipse(int rx, int ry) { reset (rx, ry); }

        void reset(int rx, int ry)
            { e.reset(rx, ry); end = false; oldx = -1; read(); }
        void reset()
            { e.reset(); end = false; oldx = -1; read(); }

        bool next()      { if (end) return false; read(); return true; }

        int x() const { return X; }
        int y() const { return Y; }
        
    private:
        ellipse e;
        bool end;
        int X, Y, oldx;

        void read();
    };

} // namespace graphics

#endif // GRAPHICS_HH
